import R from 'remeda';

/**
 * Configuration object typically passed to the constructor of the class Updux.
 */
export interface UpduxConfig<TState = any, TSubduxes = {}> {
    /**
     * Local initial state.
     * @default {}
     */
    initial?: TState;

    /**
     * Subduxes to be merged to this dux.
     */
    subduxes?: TSubduxes;
}

type StateOf<D> = D extends { initial: infer I } ? I : unknown;

export type DuxStateSubduxes<C extends {}> = keyof C extends never
    ? unknown
    : { [K in keyof C]: StateOf<C[K]> };

export class Updux<TState extends any = {}, TSubduxes = {}> {
    #localInitial: any = {};
    #subduxes;

    constructor(config: UpduxConfig<TState, TSubduxes>) {
        this.#localInitial = config.initial ?? {};
        this.#subduxes = config.subduxes ?? {};
    }

    get initial(): TState & DuxStateSubduxes<TSubduxes> {
        if (Object.keys(this.#subduxes).length === 0) return this.#localInitial;

        return Object.assign(
            {},
            this.#localInitial,
            R.mapValues(this.#subduxes, ({ initial }) => initial),
        );
    }
}
